import time

__version__ = "1.2.4"
__author__ = "Lukas Hedegaard"
__author_email__ = "lukasxhedegaard@gmail.com"
__license__ = "Apache-2.0"
__copyright__ = f'Copyright (c) 2021-{time.strftime("%Y")}, {__author__}'
__homepage__ = "https://github.com/lukashedegaard/continual-inference"
__docs__ = "A Python library for Continual Inference Networks in PyTorch"
